"""Base Class for all CalibrationInfo objects."""

import json
from collections.abc import Iterable
from dataclasses import asdict, dataclass, fields
from pathlib import Path
from typing import ClassVar, Optional, TypeVar, Union

import numpy as np
import pandas as pd
from packaging.version import Version

CalInfo = TypeVar("CalInfo", bound="CalibrationInfo")

_CAL_FORMAT_VERSION = Version("2.0.0")


@dataclass(eq=False)
class CalibrationInfo:
    """Abstract BaseClass for all Calibration Info objects.

    .. note ::
        All `CalibrationInfo` subclasses implement a `CAL_TYPE` attribute.
        If the calibration is exported into any format, this information is stored as well.
        If imported, all constructor methods intelligently infer the correct CalibrationInfo subclass based on
        this parameter.

        >>> json_string = "{cal_type: 'Ferraris', ...}"
        >>> CalibrationInfo.from_json(json_string)
        <FerrarisCalibrationInfo ...>

    """

    CAL_FORMAT_VERSION: ClassVar[Version] = _CAL_FORMAT_VERSION
    CAL_TYPE: ClassVar[str] = None
    acc_unit: Optional[str] = None
    gyr_unit: Optional[str] = None
    from_acc_unit: Optional[str] = None
    from_gyr_unit: Optional[str] = None
    comment: Optional[str] = None

    _cal_paras: ClassVar[tuple[str, ...]]

    def calibrate(
        self, acc: np.ndarray, gyr: np.ndarray, acc_unit: Optional[str], gyr_unit: Optional[str]
    ) -> tuple[np.ndarray, np.ndarray]:
        """Abstract method to perform a calibration on both acc and gyro.

        This absolutely needs to implement by any daughter class.
        It is further recommended to call `self._validate_units` in the overwritten calibrate method, to check if the
        input units are as expected.

        Parameters
        ----------
        acc :
            3D acceleration
        gyr :
            3D gyroscope values
        acc_unit
            The unit of the acceleration data
        gyr_unit
            The unit of the gyroscope data

        """
        raise NotImplementedError("This method needs to be implemented by a subclass")

    def calibrate_df(
        self,
        df: pd.DataFrame,
        acc_unit: Optional[str],
        gyr_unit: Optional[str],
        acc_cols: Iterable[str] = ("acc_x", "acc_y", "acc_z"),
        gyr_cols: Iterable[str] = ("gyr_x", "gyr_y", "gyr_z"),
    ) -> pd.DataFrame:
        """Apply the calibration to data stored in a dataframe.

        This calls `calibrate` for the respective columns and returns a copy of the df with the respective columns
        replaced by their calibrated counter-part.

        See the `calibrate` method for more information.

        Parameters
        ----------
        df :
            6 column dataframe (3 acc, 3 gyro)
        acc_cols :
            The name of the 3 acceleration columns in order x,y,z.
        gyr_cols :
            The name of the 3 acceleration columns in order x,y,z.
        acc_unit
            The unit of the acceleration data
        gyr_unit
            The unit of the gyroscope data

        Returns
        -------
        cal_df
            A copy of `df` with the calibrated data.

        """
        acc_cols = list(acc_cols)
        gyr_cols = list(gyr_cols)
        acc = df[acc_cols].to_numpy()
        gyr = df[gyr_cols].to_numpy()
        cal_acc, cal_gyr = self.calibrate(acc=acc, gyr=gyr, acc_unit=acc_unit, gyr_unit=gyr_unit)

        cal_df = df.copy()
        cal_df[acc_cols] = cal_acc
        cal_df[gyr_cols] = cal_gyr

        return cal_df

    def _validate_units(self, other_acc, other_gyr) -> None:
        check_pairs = {"acc": (other_acc, self.from_acc_unit), "gyr": (other_gyr, self.from_gyr_unit)}
        for name, (other, this) in check_pairs.items():
            if other != this:
                if this is None:
                    raise ValueError(
                        "This calibration does not provide any information about the expected input "
                        f"units for {name}. "
                        f"Set `{name}_unit` explicitly to `None` to ignore this error. "
                        "However, we recommend to recreate your calibration with proper information "
                        "about the input unit."
                    )
                raise ValueError(
                    f"The provided {name} data has a unit of {other}. "
                    f"However, the calibration is created to calibrate data with a unit of {this}."
                )

    def __eq__(self, other):
        """Check if two calibrations are identical.

        Note, we use a custom function for that, as we need to compare the numpy arrays.
        """
        # Check type:
        if not isinstance(other, self.__class__):
            raise TypeError(f"Comparison is only defined between two {self.__class__.__name__} object!")

        # Test keys equal:
        if fields(self) != fields(other):
            return False

        # Test method equal
        if not self.CAL_TYPE == other.CAL_TYPE:
            return False

        # Test all values
        for f in fields(self):
            a1 = getattr(self, f.name)
            a2 = getattr(other, f.name)
            equal = np.array_equal(a1, a2) if isinstance(a1, np.ndarray) else a1 == a2
            if not equal:
                return False
        return True

    def _to_list_dict(self):
        d = asdict(self)
        d["cal_type"] = self.CAL_TYPE
        d["_format_version"] = str(self.CAL_FORMAT_VERSION)
        return d

    @classmethod
    def _from_list_dict(cls, list_dict):
        for k in cls._cal_paras:
            list_dict[k] = np.array(list_dict[k])
        return cls(**list_dict)

    @classmethod
    def _get_subclasses(cls):
        for subclass in cls.__subclasses__():
            yield from subclass._get_subclasses()
            yield subclass

    @classmethod
    def find_subclass_from_cal_type(cls, cal_type):
        """Get a SensorCalibration subclass that handles the specified calibration type."""
        if cal_type == cls.CAL_TYPE:
            return cls
        try:
            out_cls = next(x for x in cls._get_subclasses() if cal_type == x.CAL_TYPE)
        except StopIteration as e:
            raise ValueError(
                f"No suitable calibration info class could be found for caltype `{cal_type}`. "
                f"The following classes were checked: {(cls.__name__, *(x.__name__ for x in cls._get_subclasses()))}. "
                "If your CalibrationInfo class is missing, make sure it is imported before loading a "
                "file."
            ) from e
        return out_cls

    def to_json(self) -> str:
        """Convert all calibration matrices into a json string."""
        data_dict = self._to_list_dict()
        return json.dumps(data_dict, indent=4, cls=NumpyEncoder)

    @classmethod
    def from_json(cls: type[CalInfo], json_str: str) -> CalInfo:
        """Create a calibration object from a json string (created by `CalibrationInfo.to_json`).

        Parameters
        ----------
        json_str :
            valid json string object

        Returns
        -------
        cal_info
            A CalibrationInfo object.
            The exact child class is determined by the `cal_type` key in the json string.

        """
        raw_json = json.loads(json_str)
        check_cal_format_version(Version(raw_json.pop("_format_version", None)), cls.CAL_FORMAT_VERSION)
        subclass = cls.find_subclass_from_cal_type(raw_json.pop("cal_type"))
        return subclass._from_list_dict(raw_json)

    def to_json_file(self, path: Union[str, Path]):
        """Dump acc calibration matrices into a file in json format.

        Parameters
        ----------
        path :
            path to the json file

        """
        data_dict = self._to_list_dict()
        with Path(path).open("w", encoding="utf8") as f:
            json.dump(data_dict, f, cls=NumpyEncoder, indent=4)

    @classmethod
    def from_json_file(cls: type[CalInfo], path: Union[str, Path]) -> CalInfo:
        """Create a calibration object from a valid json file (created by `CalibrationInfo.to_json_file`).

        Parameters
        ----------
        path :
            Path to the json file

        Returns
        -------
        cal_info
            A CalibrationInfo object.
            The exact child class is determined by the `cal_type` key in the json string.

        """
        with Path(path).open(encoding="utf8") as f:
            raw_json = json.load(f)
        check_cal_format_version(raw_json.pop("_format_version", None), cls.CAL_FORMAT_VERSION)
        subclass = cls.find_subclass_from_cal_type(raw_json.pop("cal_type"))
        return subclass._from_list_dict(raw_json)

    def to_hdf5(self, path: Union[str, Path]) -> None:
        """Save calibration matrices to hdf5 file format.

        Parameters
        ----------
        path :
            Path to the hdf5 file

        """
        import h5py

        with h5py.File(path, "w") as hdf:
            for k, v in self._to_list_dict().items():
                if k in self._cal_paras:
                    hdf.create_dataset(k, data=v.tolist())
                elif v:
                    hdf[k] = v

    @classmethod
    def from_hdf5(cls: type[CalInfo], path: Union[str, Path]):
        """Read calibration data stored in hdf5 fileformat (created by `CalibrationInfo.save_to_hdf5`).

        Parameters
        ----------
        path :
            Path to the hdf5 file

        Returns
        -------
        cal_info
            A CalibrationInfo object.
            The exact child class is determined by the `cal_type` key in the json string.

        """
        import h5py

        with h5py.File(path, "r") as hdf:
            format_version = hdf.get("_format_version")
            if format_version:
                format_version = format_version.asstr()[()]
            check_cal_format_version(format_version, cls.CAL_FORMAT_VERSION)
            subcls = cls.find_subclass_from_cal_type(hdf["cal_type"][()].decode("utf-8"))
            data = {}
            for k in fields(subcls):
                dp = hdf.get(k.name)
                if h5py.check_string_dtype(dp.dtype):
                    # String data
                    data[k.name] = dp.asstr()[()]
                else:
                    # No string data
                    data[k.name] = dp[()]
        return subcls._from_list_dict(data)


class NumpyEncoder(json.JSONEncoder):
    """Custom encoder for numpy array."""

    def default(self, obj):
        """Allow encoding of numpy objects by converting them to lists."""
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def check_cal_format_version(version: Optional[Version] = None, current_version: Version = _CAL_FORMAT_VERSION) -> None:
    """Check if a calibration can be loaded with the current loader."""
    # No version means, the old 1.0 format is used that does not provide a version string
    if not version:
        version = Version("1.0.0")
    if isinstance(version, str):
        version = Version(version)

    if version == current_version:
        return
    if version > current_version:
        raise ValueError("The provided version, is larger than the currently supported version.")
    if version < current_version:
        raise ValueError(
            "The provided calibration format is no longer supported. "
            "Check `imucal.legacy` if conversion helper exist."
        )



"""Calculate a Ferraris calibration from sensor data."""

from collections.abc import Iterable
from typing import ClassVar, NamedTuple, Optional, TypeVar

import numpy as np
import pandas as pd
from numpy.linalg import inv

from imucal.calibration_info import CalibrationInfo
from imucal.ferraris_calibration_info import FerrarisCalibrationInfo, TurntableCalibrationInfo

T = TypeVar("T", bound="FerrarisCalibration")


def _convert_data_from_section_list_to_df(data: pd.DataFrame, section_list: pd.DataFrame) -> pd.DataFrame:
    out = {}

    for label, row in section_list.iterrows():
        out[label] = data.iloc[row.start : row.end]

    return pd.concat(out)


class FerrarisSignalRegions(NamedTuple):
    """NamedTuple containing all signal regions required for a Ferraris Calibration."""

    acc_x_p: np.ndarray
    acc_x_a: np.ndarray
    acc_y_p: np.ndarray
    acc_y_a: np.ndarray
    acc_z_p: np.ndarray
    acc_z_a: np.ndarray
    gyr_x_p: np.ndarray
    gyr_x_a: np.ndarray
    gyr_y_p: np.ndarray
    gyr_y_a: np.ndarray
    gyr_z_p: np.ndarray
    gyr_z_a: np.ndarray

    acc_x_rot: np.ndarray
    acc_y_rot: np.ndarray
    acc_z_rot: np.ndarray
    gyr_x_rot: np.ndarray
    gyr_y_rot: np.ndarray
    gyr_z_rot: np.ndarray

    def validate(self) -> None:
        """Validate that all fields are populated with numpy arrays."""
        for k in self._fields:
            if not isinstance(getattr(self, k), np.ndarray) or len(getattr(self, k)) == 0:
                raise ValueError("The the signal region {} is no valid numpy array.")


class FerrarisCalibration:
    """Calculate a Ferraris calibration matrices based on a set of calibration movements.

    The Ferraris calibration is derived based on a well defined series of data recordings:

    =====  ======================================================================
    Static Holds
    -----------------------------------------------------------------------------
    Name   Explanation
    =====  ======================================================================
    `x_p`  positive x-axis of sensor is aligned with gravity (x-acc measures +1g)
    `x_a`  negative x-axis of sensor is aligned with gravity (x-acc measures -1g)
    `y_p`  positive y-axis of sensor is aligned with gravity (y-acc measures +1g)
    `y_a`  negative y-axis of sensor is aligned with gravity (y-acc measures -1g)
    `z_p`  positive z-axis of sensor is aligned with gravity (z-acc measures +1g)
    `z_a`  negative z-axis of sensor is aligned with gravity (z-acc measures -1g)
    =====  ======================================================================

    =======  ========================================================================================================
    Rotations
    -----------------------------------------------------------------------------------------------------------------
    Name     Explanation
    =======  ========================================================================================================
    `x_rot`  sensor is rotated clockwise in the `x_p` position around the x-axis (x-gyro shows negative values) for a
             well known angle (typically 360 deg)
    `y_rot`  sensor is rotated clockwise in the `y_p` position around the y-axis (y-gyro shows negative values) for a
             well known angle (typically 360 deg)
    `z_rot`  sensor is rotated clockwise in the `z_p` position around the z-axis (z-gyro shows negative values) for a
             well known angle (typically 360 deg)
    =======  ========================================================================================================

    All sections need to be recorded for a sensor and then annotated.
    This class then takes the data of each section (represented as a
    :class:`~imucal.FerrarisSignalRegions` object) and calculates the calibration matrizes for
    the gyroscope and accelerometer.

    As it is quite tedious to obtain the data of each section in a seperate array, you should make use of the available
    helper functions to turn a continous recording into annotated sections (See the See also section)

    Parameters
    ----------
    sampling_rate :
        Sampling rate of the data
    expected_angle :
        expected rotation angle for the gyroscope rotation.
    grav :
        The expected value of the gravitational acceleration.
    calibration_info_class :
        The calibration Info class to use to store the final calibration information.
        This should be a FerrarisCalibrationInfo or a custom subclass.

    Notes
    -----
    Depending on how the axis of your respective sensor coordinate system are defined and how you perform the
    calibration, you might need to change the `grav` and `expected_angle` parameter.

    Typical situations are:

    - If you define the positive axis direction as the direction, where the acc measures -g, change `grav` to
      -9.81 m/s^2
    - If you perform a counter-clockwise rotation during the calibration, set `expected_angle` to +360
    - For combinations of both, both parameter might need to be adapted


    Examples
    --------
    >>> from imucal import FerrarisCalibration, ferraris_regions_from_interactive_plot
    >>> sampling_rate = 100  # Hz
    >>> data = ...  # my data as 6 col pandas dataframe
    >>> # This will open an interactive plot, where you can select the start and the stop sample of each region
    >>> section_data, section_list = ferraris_regions_from_interactive_plot(
    ...     data, sampling_rate=sampling_rate
    ... )
    >>> section_list.to_csv(
    ...     "./calibration_sections.csv"
    ... )  # Store the annotated section list as reference for the
    ... # future
    >>> cal = FerrarisCalibration()  # Create new calibration object
    >>> calibration_info = cal.compute(  # Calculate the actual matrizes.
    ...     section_data,
    ...     sampling_rate_hz=sampling_rate,
    ...     from_acc_unit="a.u.",
    ...     from_gyr_unit="a.u.",
    ...     comment="my custom comment."
    ...)
    >>> calibration_info_class
    < FerrarisCalibration object at ... >

    See Also
    --------
    imucal.ferraris_regions_from_df: Generate valid sections from preannotated dataframe.
    imucal.ferraris_regions_from_interactive_plot: Generate valid sections via manual annotation in an interactive
        GUI.
    imucal.ferraris_regions_from_section_list: Generate valid sections based on raw data and start-end labels for the
        individual sections.

    """

    grav: float
    expected_angle: float
    calibration_info_class: type[FerrarisCalibrationInfo]

    OUT_ACC_UNIT: ClassVar[str] = "m/s^2"
    OUT_GYR_UNIT: ClassVar[str] = "deg/s"

    FERRARIS_SECTIONS: ClassVar[tuple[str, ...]] = (
        "x_p",
        "x_a",
        "y_p",
        "y_a",
        "z_p",
        "z_a",
        "x_rot",
        "y_rot",
        "z_rot",
    )

    def __init__(
        self,
        grav: float = 9.81,
        expected_angle: float = -360,
        calibration_info_class: type[FerrarisCalibrationInfo] = FerrarisCalibrationInfo,
    ) -> None:
        self.grav = grav
        self.expected_angle = expected_angle
        self.calibration_info_class = calibration_info_class

    def compute(
        self,
        signal_regions: FerrarisSignalRegions,
        sampling_rate_hz: float,
        from_acc_unit: str,
        from_gyr_unit: str,
        **kwargs,
    ) -> CalibrationInfo:
        """Compute the calibration Information.

        This actually performs the Ferraris calibration following the original publication equation by equation.
        """
        signal_regions.validate()

        # Initialize the cal info with all the meta data
        cal_mat = self.calibration_info_class(
            from_acc_unit=from_acc_unit,
            from_gyr_unit=from_gyr_unit,
            acc_unit=self.OUT_ACC_UNIT,
            gyr_unit=self.OUT_GYR_UNIT,
            **kwargs,
        )

        ###############################################################################################################
        # Compute Acceleration Matrix

        # Calculate means from all static phases and stack them into 3x3 matrices
        # Note: Each measurement should be a column
        U_a_p = np.vstack(  # noqa: N806
            (
                np.mean(signal_regions.acc_x_p, axis=0),
                np.mean(signal_regions.acc_y_p, axis=0),
                np.mean(signal_regions.acc_z_p, axis=0),
            )
        ).T
        U_a_n = np.vstack(  # noqa: N806
            (
                np.mean(signal_regions.acc_x_a, axis=0),
                np.mean(signal_regions.acc_y_a, axis=0),
                np.mean(signal_regions.acc_z_a, axis=0),
            )
        ).T

        # Eq. 19
        U_a_s = U_a_p + U_a_n  # noqa: N806

        # Bias Matrix
        # Eq. 20
        B_a = U_a_s / 2  # noqa: N806

        # Bias Vector
        b_a = np.diag(B_a)
        cal_mat.b_a = b_a

        # Compute Scaling and Rotation
        # No need for bias correction, since it cancels out!
        # Eq. 21
        U_a_d = U_a_p - U_a_n  # noqa: N806

        # Calculate Scaling matrix
        # Eq. 23
        k_a_sq = 1 / (4 * self.grav**2) * np.diag(U_a_d @ U_a_d.T)
        K_a = np.diag(np.sqrt(k_a_sq))  # noqa: N806
        cal_mat.K_a = K_a

        # Calculate Rotation matrix
        # Eq. 22
        R_a = inv(K_a) @ U_a_d / (2 * self.grav)  # noqa: N806
        cal_mat.R_a = R_a

        ###############################################################################################################
        # Calculate Gyroscope Matrix

        # Gyro Bias from the static phases of the acc calibration
        # One static phase would be sufficient, but why not use all of them if you have them.
        # Note that this calibration ignores any influences due to the earth rotation.

        b_g = np.mean(
            np.vstack(
                (
                    signal_regions.gyr_x_p,
                    signal_regions.gyr_x_a,
                    signal_regions.gyr_y_p,
                    signal_regions.gyr_y_a,
                    signal_regions.gyr_z_p,
                    signal_regions.gyr_z_a,
                )
            ),
            axis=0,
        )

        cal_mat.b_g = b_g

        # Acceleration sensitivity

        # Note: Each measurement should be a column
        U_g_p = np.vstack(  # noqa: N806
            (
                np.mean(signal_regions.gyr_x_p, axis=0),
                np.mean(signal_regions.gyr_y_p, axis=0),
                np.mean(signal_regions.gyr_z_p, axis=0),
            )
        ).T
        U_g_a = np.vstack(  # noqa: N806
            (
                np.mean(signal_regions.gyr_x_a, axis=0),
                np.mean(signal_regions.gyr_y_a, axis=0),
                np.mean(signal_regions.gyr_z_a, axis=0),
            )
        ).T

        # Eq. 9
        K_ga = (U_g_p - U_g_a) / (2 * self.grav)  # noqa: N806
        cal_mat.K_ga = K_ga

        # Gyroscope Scaling and Rotation

        # First apply partial calibration to remove offset and acc influence
        acc_x_rot_cor = cal_mat._calibrate_acc(signal_regions.acc_x_rot)
        acc_y_rot_cor = cal_mat._calibrate_acc(signal_regions.acc_y_rot)
        acc_z_rot_cor = cal_mat._calibrate_acc(signal_regions.acc_z_rot)
        gyr_x_rot_cor = cal_mat._calibrate_gyr_offsets(signal_regions.gyr_x_rot, acc_x_rot_cor)
        gyr_y_rot_cor = cal_mat._calibrate_gyr_offsets(signal_regions.gyr_y_rot, acc_y_rot_cor)
        gyr_z_rot_cor = cal_mat._calibrate_gyr_offsets(signal_regions.gyr_z_rot, acc_z_rot_cor)

        # Integrate gyro readings
        # Eg. 13/14
        W_s = np.zeros((3, 3))  # noqa: N806
        W_s[:, 0] = np.sum(gyr_x_rot_cor, axis=0) / sampling_rate_hz
        W_s[:, 1] = np.sum(gyr_y_rot_cor, axis=0) / sampling_rate_hz
        W_s[:, 2] = np.sum(gyr_z_rot_cor, axis=0) / sampling_rate_hz

        # Eq.15
        expected_angles = self.expected_angle * np.identity(3)
        multiplied = W_s @ inv(expected_angles)

        # Eq. 12
        k_g_sq = np.diag(multiplied @ multiplied.T)
        K_g = np.diag(np.sqrt(k_g_sq))  # noqa: N806
        cal_mat.K_g = K_g

        R_g = inv(K_g) @ multiplied  # noqa: N806
        cal_mat.R_g = R_g

        return cal_mat


class TurntableCalibration(FerrarisCalibration):
    """Calculate a Ferraris calibration matrices based on a turntable measurement.

    This calibration is basically identical to the FerrarisCalibration.
    However, the calibrate method will return a :class:`~imucal.TurntableCalibrationInfo` to indicate the expected
    higher precision of this calibration method.

    Further this Calibration expects rotations of 720 deg by default, as this is common for many turntables.
    For further information on the sign of the expected rotation angle see the :class:`~imucal.FerrarisCalibration`.

    See Also
    --------
    imucal.FerrarisCalibration

    """

    def __init__(
        self,
        grav: float = 9.81,
        expected_angle: float = -720,
        calibration_info_class: type[TurntableCalibrationInfo] = TurntableCalibrationInfo,
    ) -> None:
        super().__init__(grav=grav, expected_angle=expected_angle, calibration_info_class=calibration_info_class)


def ferraris_regions_from_df(
    df: pd.DataFrame,
    acc_cols: Optional[Iterable[str]] = ("acc_x", "acc_y", "acc_z"),
    gyr_cols: Optional[Iterable[str]] = ("gyr_x", "gyr_y", "gyr_z"),
) -> FerrarisSignalRegions:
    """Create a Calibration object based on a dataframe which has all required sections labeled.

    The expected Dataframe has the section label as index and has at least the 6 required data columns.
    The index must contain all the sections listed in :class:`~imucal.FerrarisCalibration``.FERRARIS_SECTIONS`.

    Examples
    --------
    >>> import pandas as pd
    >>> sampling_rate = 100  # Hz
    >>> df = ...  # A valid DataFrame with all sections in the index
    >>> print(df)
            acc_x acc_y   acc_z  gyr_x  gyr_y  gyr_z
    part
    x_a   -2052.0 -28.0   -73.0    1.0    0.0   -5.0
    x_a   -2059.0 -29.0   -77.0    2.0   -3.0   -5.0
    x_a   -2054.0 -25.0   -71.0    3.0   -2.0   -3.0
    ...       ...   ...     ...    ...    ...    ...
    z_rot   -36.0  35.0  2079.0    2.0   -5.0   -2.0
    z_rot   -28.0  36.0  2092.0    6.0   -5.0   -4.0
    z_rot   -36.0  21.0  2085.0    5.0   -4.0   -4.0
    >>> regions = ferraris_regions_from_df(df)
    >>> regions
    FerrarisSignalRegions(x_a=array([...]), ..., z_rot=array([...]))

    Parameters
    ----------
    df :
        6 column dataframe (3 acc, 3 gyro)
    acc_cols :
        The name of the 3 acceleration columns in order x,y,z.
    gyr_cols :
        The name of the 3 acceleration columns in order x,y,z.

    Returns
    -------
    ferraris_cal_obj : FerrarisSignalRegions

    See Also
    --------
    ferraris_regions_from_interactive_plot
    ferraris_regions_from_section_list

    """
    acc_df = df[list(acc_cols)]
    gyro_df = df[list(gyr_cols)]
    acc_dict = acc_df.groupby(level=0).apply(lambda x: x.to_numpy()).to_dict()
    gyro_dict = gyro_df.groupby(level=0).apply(lambda x: x.to_numpy()).to_dict()
    acc_dict = {"acc_" + k: v for k, v in acc_dict.items()}
    gyro_dict = {"gyr_" + k: v for k, v in gyro_dict.items()}

    return FerrarisSignalRegions(**acc_dict, **gyro_dict)


def ferraris_regions_from_section_list(
    data: pd.DataFrame,
    section_list: pd.DataFrame,
    acc_cols: Optional[Iterable[str]] = ("acc_x", "acc_y", "acc_z"),
    gyr_cols: Optional[Iterable[str]] = ("gyr_x", "gyr_y", "gyr_z"),
) -> FerrarisSignalRegions:
    """Create a Calibration object based on a valid section list.

    A section list marks the start and the endpoints of each required section in the data object.
    A valid section list is usually created using `FerrarisCalibration.from_interactive_plot()`.
    This section list can be stored on disk and this method can be used to turn it back into a valid calibration
    object.

    Parameters
    ----------
    data :
        6 column dataframe (3 acc, 3 gyro)
    section_list :
        A pandas dataframe representing a section list
    acc_cols :
        The name of the 3 acceleration columns in order x,y,z.
        Defaults to `FerrarisCalibration.ACC_COLS`
    gyr_cols :
        The name of the 3 acceleration columns in order x,y,z.
        Defaults to `FerrarisCalibration.GYRO_COLS`

    Returns
    -------
    ferraris_cal_obj : FerrarisSignalRegions


    Examples
    --------
    >>> import pandas as pd
    >>> # Load a valid section list from disk. Note the `index_col=0` to preserve correct format!
    >>> section_list = pd.read_csv("./calibration_sections.csv", index_col=0)
    >>> sampling_rate = 100  # Hz
    >>> df = ...  # my data as 6 col pandas dataframe
    >>> regions = ferraris_regions_from_section_list(df)
    >>> regions
    FerrarisSignalRegions(x_a=array([...]), ..., z_rot=array([...]))

    See Also
    --------
    ferraris_regions_from_interactive_plot
    ferraris_regions_from_df

    """
    df = _convert_data_from_section_list_to_df(data, section_list)
    return ferraris_regions_from_df(df, acc_cols=acc_cols, gyr_cols=gyr_cols)


def ferraris_regions_from_interactive_plot(
    data: pd.DataFrame,
    acc_cols: Iterable[str] = ("acc_x", "acc_y", "acc_z"),
    gyr_cols: Iterable[str] = ("gyr_x", "gyr_y", "gyr_z"),
    title: Optional[str] = None,
    figsize=(20, 10),
) -> tuple[FerrarisSignalRegions, pd.DataFrame]:
    """Create a Calibration object by selecting the individual signal sections manually in an interactive GUI.

    This will open a Tkinter Window that allows you to label the start and the end all required sections for a
    Ferraris Calibration.
    See the class docstring for more detailed explanations of these sections.

    Examples
    --------
    >>> sampling_rate = 100  # Hz
    >>> data = ...  # my data as 6 col pandas dataframe
    >>> # This will open an interactive plot, where you can select the start and the stop sample of each region
    >>> regions, section_list = ferraris_regions_from_interactive_plot(
    ...     data, sampling_rate=sampling_rate
    ... )
    >>> section_list.to_csv(
    ...     "./calibration_sections.csv"
    ... )  # This is optional, but recommended
    >>> regions
    FerrarisSignalRegions(x_a=array([...]), ..., z_rot=array([...]))

    Parameters
    ----------
    data :
        6 column dataframe (3 acc, 3 gyro)
    acc_cols :
        The name of the 3 acceleration columns in order x,y,z.
    gyr_cols :
        The name of the 3 acceleration columns in order x,y,z.
    title :
        Optional title of the plot window
    figsize :
        The initial size of the plot window.
        If you can not see the toolbar at the bottom, it might help to reduce the figure size.

    Returns
    -------
    ferraris_cal_obj : FerrarisSignalRegions
    section_list : pd.DataFrame
        Section list representing the start and stop of each section.
        It is advised to save this to disk to avoid repeated manual labeling.
        :py:func:`~imucal.ferraris_regions_from_section_list` can be used to recreate the regions object

    See Also
    --------
    ferraris_regions_from_section_list
    ferraris_regions_from_df

    """
    acc = data[list(acc_cols)].to_numpy()
    gyr = data[list(gyr_cols)].to_numpy()

    section_list = _find_ferraris_regions_interactive(acc, gyr, title=title, figsize=figsize)
    return (
        ferraris_regions_from_section_list(data, section_list, gyr_cols=gyr_cols, acc_cols=acc_cols),
        section_list,
    )


def _find_ferraris_regions_interactive(
    acc: np.ndarray, gyro: np.ndarray, title: Optional[str] = None, figsize=(20, 10)
):
    """Prepare the calibration data for the later calculation of calibration matrices.

    Parameters
    ----------
    acc : (n, 3) array
        Acceleration data
    gyro : (n, 3) array
        Gyroscope data
    title :
        Optional title for the Calibration GUI


    """
    from imucal.calibration_gui import CalibrationGui

    plot = CalibrationGui(acc, gyro, FerrarisCalibration.FERRARIS_SECTIONS, title=title, initial_figsize=figsize)

    section_list = plot.section_list

    check_all = (all(v) for v in section_list.values())
    if not all(check_all):
        raise ValueError("Some regions are missing in the section list. Label all regions before closing the plot")

    section_list = pd.DataFrame(section_list, index=("start", "end")).T

    return section_list



"""Wrapper object to hold calibration matrices for a Ferraris Calibration."""

from dataclasses import dataclass
from typing import ClassVar, Optional

import numpy as np

from imucal.calibration_info import CalibrationInfo


@dataclass(eq=False)
class FerrarisCalibrationInfo(CalibrationInfo):
    """Calibration object that represents all the required information to apply a Ferraris calibration to a dataset.

    Parameters
    ----------
    K_a :
        Scaling matrix for the acceleration
    R_a :
        Rotation matrix for the acceleration
    b_a :
        Acceleration bias
    K_g :
        Scaling matrix for the gyroscope
    R_g :
        Rotation matrix for the gyroscope
    K_ga :
        Influence of acceleration on gyroscope
    b_g :
        Gyroscope bias

    """

    CAL_TYPE: ClassVar[str] = "Ferraris"

    acc_unit: str = "m/s^2"
    gyr_unit: str = "deg/s"
    K_a: Optional[np.ndarray] = None
    R_a: Optional[np.ndarray] = None
    b_a: Optional[np.ndarray] = None
    K_g: Optional[np.ndarray] = None
    R_g: Optional[np.ndarray] = None
    K_ga: Optional[np.ndarray] = None
    b_g: Optional[np.ndarray] = None

    _cal_paras: ClassVar[tuple[str, ...]] = ("K_a", "R_a", "b_a", "K_g", "R_g", "K_ga", "b_g")

    def calibrate(
        self, acc: np.ndarray, gyr: np.ndarray, acc_unit: Optional[str], gyr_unit: Optional[str]
    ) -> tuple[np.ndarray, np.ndarray]:
        """Calibrate the accelerometer and the gyroscope.

        This corrects:
            acc: scaling, rotation, non-orthogonalities, and bias
            gyro: scaling, rotation, non-orthogonalities, bias, and acc influence on gyro

        Parameters
        ----------
        acc :
            3D acceleration
        gyr :
            3D gyroscope values
        acc_unit
            The unit of the acceleration data
        gyr_unit
            The unit of the gyroscope data

        Returns
        -------
        Corrected acceleration and gyroscope values

        """
        # Check if all required paras are initialized to throw appropriate error messages:
        for v in self._cal_paras:
            if getattr(self, v, None) is None:
                raise ValueError(
                    f"{self._cal_paras} need to initialised before an acc calibration can be performed. {v} is missing."
                )
        self._validate_units(acc_unit, gyr_unit)
        acc_out = self._calibrate_acc(acc)
        gyro_out = self._calibrate_gyr(gyr, acc_out)

        return acc_out, gyro_out

    def _calibrate_acc(self, acc: np.ndarray) -> np.ndarray:
        """Calibrate the accelerometer.

        This corrects scaling, rotation, non-orthogonalities and bias.

        Parameters
        ----------
        acc :
            3D acceleration

        Returns
        -------
        Calibrated acceleration

        """
        # Check if all required paras are initialized to throw appropriate error messages:
        paras = ("K_a", "R_a", "b_a")
        for v in paras:
            if getattr(self, v, None) is None:
                raise ValueError(
                    f"{paras} need to initialised before an acc calibration can be performed. {v} is missing"
                )

        # Combine Scaling and rotation matrix to one matrix
        acc_mat = np.linalg.inv(self.R_a) @ np.linalg.inv(self.K_a)
        acc_out = acc_mat @ (acc - self.b_a).T

        return acc_out.T

    def _calibrate_gyr(self, gyr, calibrated_acc=None):
        # Check if all required paras are initialized to throw appropriate error messages:
        required = ["K_g", "R_g", "b_g"]
        if calibrated_acc is not None:
            required += ["K_ga"]
        for v in required:
            if getattr(self, v, None) is None:
                raise ValueError(
                    f"{required} need to initialised before an gyro calibration can be performed. {v} is missing"
                )
        # Combine Scaling and rotation matrix to one matrix
        gyro_mat = np.matmul(np.linalg.inv(self.R_g), np.linalg.inv(self.K_g))
        tmp = self._calibrate_gyr_offsets(gyr, calibrated_acc)

        gyro_out = gyro_mat @ tmp.T
        return gyro_out.T

    def _calibrate_gyr_offsets(self, gyr, calibrated_acc=None):
        d_ga = np.array(0) if calibrated_acc is None else self.K_ga @ calibrated_acc.T
        offsets = d_ga.T + self.b_g
        return gyr - offsets


class TurntableCalibrationInfo(FerrarisCalibrationInfo):
    """Calibration object that represents all the required information to apply a Turntable calibration to a dataset.

    A Turntable calibration is identical to a Ferraris calibration.
    However, because the parameters are calculated using a calibration table instead of hand rotations,
    higher precision is expected.

    Parameters
    ----------
    K_a :
        Scaling matrix for the acceleration
    R_a :
        Rotation matrix for the acceleration
    b_a :
        Acceleration bias
    K_g :
        Scaling matrix for the gyroscope
    R_g :
        Rotation matrix for the gyroscope
    K_ga :
        Influence of acceleration on gyroscope
    b_g :
        Gyroscope bias

    """

    CAL_TYPE = "Turntable"
